from __future__ import annotations

import json
from typing import Callable

from palimpzest.constants import AggFunc, Cardinality
from palimpzest.core.data.datasources import DataSource
from palimpzest.core.elements.filters import Filter
from palimpzest.core.elements.groupbysig import GroupBySig
from palimpzest.core.lib.schemas import Number, Schema
from palimpzest.datamanager.datamanager import DataDirectory
from palimpzest.utils.hash_helpers import hash_for_id
from palimpzest.utils.index_helpers import get_index_str


#####################################################
#
#####################################################
class Set:
    """
    A Set is the logical abstraction for a set of DataRecords matching some Schema. It is
    also a node in the computation graph of a Dataset.

    Each Dataset consists of one or more Sets. The "initial" Set in a Dataset can be thought
    of as the Set that results from reading each DataRecord unaltered from the source. For each
    filter or transformation that is applied to the Dataset, a new Set is created which defines
    the set of DataRecords that result from applying that filter or transformation. In brief,
    the Sets define a Dataset's computation graph. Sets can also be cached to maximize the reuse
    of past computation.

    Sets are initialized with a dataset_id, a schema, and a source. The source is either an
    existing Set or a raw data source (such as a directory or S3 prefix). Sets may be initialized
    with a Filter (which defines the filtering performed on the source to obtain *this* Set),
    and a description of what this Set is meant to represent.
    """

    SET_VERSION = 0.1

    def __init__(
        self,
        source: Set | DataSource,
        schema: Schema,
        desc: str | None = None,
        filter: Filter | None = None,
        udf: Callable | None = None,
        agg_func: AggFunc | None = None,
        group_by: GroupBySig | None = None,
        project_cols: list[str] | None = None,
        index = None, # TODO(Siva): Abstract Index and add a type here and elsewhere
        search_attr: str | None = None,
        output_attr: str | None = None,
        k: int | None = None, # TODO: disambiguate `k` to be something like `retrieve_k`
        limit: int | None = None,
        cardinality: Cardinality = Cardinality.ONE_TO_ONE,
        depends_on: list[str] | None = None,
        nocache: bool = False,
    ):
        self._schema = schema
        self._source = source
        self._desc = desc
        self._filter = filter
        self._udf = udf
        self._agg_func = agg_func
        self._group_by = group_by
        self._project_cols = None if project_cols is None else sorted(project_cols)
        self._index = index
        self._search_attr = search_attr
        self._output_attr = output_attr
        self._k = k
        self._limit = limit
        self._cardinality = cardinality
        self._depends_on = [] if depends_on is None else sorted(depends_on)
        self._nocache = nocache

    def __str__(self):
        return (
            f"{self.__class__.__name__}(schema={self.schema}, desc={self._desc}, "
            f"filter={str(self._filter)}, udf={str(self._udf)}, agg_func={str(self._agg_func)}, limit={str(self._limit)}, "
            f"project_cols={str(self._project_cols)}, uid={self.universal_identifier()})"
        )

    @property
    def schema(self) -> Schema:
        return self._schema

    def serialize(self):
        # NOTE: I needed to remove depends_on from the serialization dictionary because
        # the optimizer changes the name of the depends_on fields to be their "full" name.
        # This created an issue with the node.universal_identifier() not being consistent
        # after changing the field to its full name.
        d = {
            "version": Set.SET_VERSION,
            "schema": self.schema.json_schema(),
            "source": self._source.serialize(),
            "desc": repr(self._desc),
            "filter": None if self._filter is None else self._filter.serialize(),
            "udf": None if self._udf is None else str(self._udf),
            "agg_func": None if self._agg_func is None else self._agg_func.value,
            "cardinality": self._cardinality,
            "limit": self._limit,
            "group_by": (None if self._group_by is None else self._group_by.serialize()),
            "project_cols": (None if self._project_cols is None else self._project_cols),
            "index": None if self._index is None else get_index_str(self._index),
            "search_attr": self._search_attr,
            "output_attr": self._output_attr,
            "k": self._k,
        }

        return d

    def universal_identifier(self):
        """Return a unique identifier for this Set."""
        d = self.serialize()
        ordered = json.dumps(d, sort_keys=True)
        result = hash_for_id(ordered)
        return result

    def json_schema(self):
        """Return the JSON schema for this Set."""
        return self.schema.json_schema()


class Dataset(Set):
    """
    A Dataset is the intended abstraction for programmers to interact with when manipulating Sets.

    Users instantiate a Dataset by specifying a `source` that either points to a
    DataSource or an existing cached Set. Users can then perform computations on
    the Dataset in an imperative fashion by leveraging functions such as `filter`,
    `convert`, `aggregate`, etc. Underneath the hood, each of these operations creates
    a new Set which is cached by the DataManager. As a result, the Sets define the
    lineage of computation on a Dataset, and this enables programmers to re-use
    previously cached computation by providing it as a `source` to some future Dataset.
    """

    def __init__(self, source: str | DataSource, *args, **kwargs):
        # convert source (str) -> source (DataSource) if need be
        source = DataDirectory().get_registered_dataset(source) if isinstance(source, str) else source

        # intialize class
        super().__init__(source, *args, **kwargs)

    def copy(self) -> Dataset:
        source_copy = self._source.copy()
        dataset_copy = Dataset(
            schema=self.schema,
            source=source_copy,
            desc=self._desc,
            filter=self._filter,
            udf=self._udf,
            agg_func=self._agg_func,
            group_by=self._group_by,
            index=self._index,
            search_attr=self._search_attr,
            output_attr=self._output_attr,
            k=self._k,
            limit=self._limit,
            cardinality=self._cardinality,
            depends_on=self._depends_on,
            nocache=self._nocache,
        )
        return dataset_copy

    def filter(
        self,
        _filter: str | Callable,
        depends_on: str | list[str] | None = None,
    ) -> Dataset:
        """Add a filter to the Set. This filter will possibly restrict the items that are returned later."""
        f = None
        if isinstance(_filter, str):
            f = Filter(_filter)
        elif callable(_filter):
            f = Filter(filter_fn=_filter)
        else:
            raise Exception("Filter type not supported.", type(_filter))

        if isinstance(depends_on, str):
            depends_on = [depends_on]

        return Dataset(
            source=self,
            schema=self.schema,
            filter=f,
            depends_on=depends_on,
            nocache=self._nocache,
        )

    def convert(
        self,
        output_schema: Schema,
        udf: Callable | None = None,
        cardinality: Cardinality = Cardinality.ONE_TO_ONE,
        depends_on: str | list[str] | None = None,
        desc: str = "Convert to new schema",
    ) -> Dataset:
        """Convert the Set to a new schema."""
        if isinstance(depends_on, str):
            depends_on = [depends_on]

        return Dataset(
            source=self,
            schema=output_schema,
            udf=udf,
            cardinality=cardinality,
            depends_on=depends_on,
            desc=desc,
            nocache=self._nocache,
        )

    def count(self) -> Dataset:
        """Apply a count aggregation to this set"""
        return Dataset(
            source=self,
            schema=Number,
            desc="Count results",
            agg_func=AggFunc.COUNT,
            nocache=self._nocache,
        )

    def average(self) -> Dataset:
        """Apply an average aggregation to this set"""
        return Dataset(
            source=self,
            schema=Number,
            desc="Average results",
            agg_func=AggFunc.AVERAGE,
            nocache=self._nocache,
        )

    def groupby(self, groupby: GroupBySig) -> Dataset:
        return Dataset(
            source=self,
            schema=groupby.output_schema(),
            desc="Group By",
            group_by=groupby,
            nocache=self._nocache,
        )

    def retrieve(self, output_schema, index, search_attr, output_attr, k=-1) -> Dataset:
        return Dataset(
            source=self,
            schema=output_schema,
            desc="Retrieve",
            index=index,
            search_attr=search_attr,
            output_attr=output_attr,
            k=k,
            nocache=self._nocache,
        )

    def limit(self, n: int) -> Dataset:
        """Limit the set size to no more than n rows"""
        return Dataset(
            source=self,
            schema=self.schema,
            desc="LIMIT " + str(n),
            limit=n,
            nocache=self._nocache,
        )

    def project(self, project_cols: list[str] | str) -> Dataset:
        """Project the Set to only include the specified columns."""
        return Dataset(
            source=self,
            schema=self.schema.project(project_cols),
            project_cols=project_cols if isinstance(project_cols, list) else [project_cols],
            nocache=self._nocache,
        )



from palimpzest.core.data.datasources import (
    DataSource,
    DirectorySource,
    FileSource,
    HTMLFileDirectorySource,
    ImageFileDirectorySource,
    MemorySource,
    PDFFileDirectorySource,
    TextFileDirectorySource,
    UserSource,
    ValidationDataSource,
    XLSFileDirectorySource,
)
from palimpzest.core.elements.records import DataRecord, DataRecordSet
from palimpzest.core.lib.fields import (
    BooleanField,
    BytesField,
    CallableField,
    Field,
    ListField,
    NumericField,
    StringField,
)
from palimpzest.core.lib.schemas import (
    URL,
    Download,
    EquationImage,
    File,
    ImageFile,
    Number,
    OperatorDerivedSchema,
    PDFFile,
    PlotImage,
    RawJSONObject,
    Schema,
    SourceRecord,
    Table,
    TextFile,
    WebPage,
    XLSFile,
)

__all__ = [
    # fields
    "BooleanField",
    "BytesField",
    "CallableField",
    "Field",
    "ListField",
    "NumericField",
    "StringField",
    # schemas
    "URL",
    "Download",
    "EquationImage",
    "File",
    "ImageFile",
    "Number",
    "OperatorDerivedSchema",
    "PDFFile",
    "PlotImage",
    "RawJSONObject",
    "Schema",
    "SourceRecord",
    "Table",
    "TextFile",
    "WebPage",
    "XLSFile",
    # datasources
    "DataSource",
    "DirectorySource",
    "FileSource",
    "HTMLFileDirectorySource",
    "ImageFileDirectorySource",
    "MemorySource",
    "PDFFileDirectorySource",
    "TextFileDirectorySource",
    "UserSource",
    "ValidationDataSource",
    "XLSFileDirectorySource",
    # records
    "DataRecord",
    "DataRecordSet",
]



from __future__ import annotations

import json
from typing import Any

import pandas as pd

from palimpzest.constants import DERIVED_SCHEMA_PREFIX, FROM_DF_PREFIX
from palimpzest.core.data.dataclasses import RecordOpStats
from palimpzest.core.lib.fields import Field
from palimpzest.core.lib.schemas import Schema
from palimpzest.utils.hash_helpers import hash_for_id, hash_for_temp_schema


class DataRecord:
    """A DataRecord is a single record of data matching some Schema."""

    def __init__(
        self,
        schema: Schema,
        source_id: int | str,
        parent_id: str | None = None,
        cardinality_idx: int | None = None,
    ):
        # check that source_id is provided
        assert source_id is not None, "Every DataRecord must be constructed with a source_id"

        # schema for the data record
        self.schema = schema

        # mapping from field names to Field objects; effectively a mapping from a field name to its type        
        self.field_types: dict[str, Field] = schema.field_map()

        # mapping from field names to their values
        self.field_values: dict[str, Any] = {}

        # the source record(s) from which this DataRecord is derived
        self.source_id = str(source_id)

        # the id of the parent record(s) from which this DataRecord is derived
        self.parent_id = parent_id

        # store the cardinality index
        self.cardinality_idx = cardinality_idx

        # indicator variable which may be flipped by filter operations to signal when a record has been filtered out
        self.passed_operator = True

        # NOTE: Record ids are hashed based on:
        # 0. their schema (keys)
        # 1. their parent record id(s) (or source_id if there is no parent record)
        # 2. their index in the fan out (if this is in a one-to-many operation)
        #
        # We currently do NOT hash just based on record content (i.e. schema (key, value) pairs)
        # because multiple outputs for a given operation may have the exact same
        # schema (key, value) pairs.
        #
        # We may revisit this hashing scheme in the future.

        # unique identifier for the record
        id_str = (
            str(schema) + (parent_id if parent_id is not None else self.source_id)
            if cardinality_idx is None
            else str(schema) + str(cardinality_idx) + (parent_id if parent_id is not None else self.source_id)
        )
        self.id = hash_for_id(id_str)


    def __setattr__(self, name: str, value: Any, /) -> None:
        if name in ["schema", "field_types", "field_values", "source_id", "parent_id", "cardinality_idx", "passed_operator", "id"]:
            super().__setattr__(name, value)
        else:
            self.field_values[name] = value


    def __getattr__(self, name: str) -> Any:
        if name == "field_values":
            pass
        elif name in self.field_values:
            return self.field_values[name]
        else:
            raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'")


    def __getitem__(self, field: str) -> Any:
        return self.__getattr__(field)


    def __setitem__(self, field: str, value: Any) -> None:
        self.__setattr__(field, value)


    def __str__(self, truncate: int | None = 15) -> str:
        if truncate is not None:
            items = (f"{k}={str(v)[:truncate]!r}{'...' if len(str(v)) > truncate else ''}" for k, v in sorted(self.field_values.items()))
        else:
            items = (f"{k}={v!r}" for k, v in sorted(self.field_values.items()))
        return "{}({})".format(type(self).__name__, ", ".join(items))

    def __repr__(self) -> str:
        return self.__str__(truncate=None)

    def __eq__(self, other):
        return isinstance(other, DataRecord) and self.field_values == other.field_values and self.schema.get_desc() == other.schema.get_desc()


    def __hash__(self):
        return hash(self.as_json_str())


    def __iter__(self):
        yield from self.field_values.items()


    def get_field_names(self):
        return list(self.field_types.keys())


    def get_field_type(self, field_name: str) -> Field:
        return self.field_types[field_name]


    def copy(self, include_bytes: bool = True, project_cols: list[str] | None = None):
        # make new record which has parent_record as its parent (and the same source_id)
        new_dr = DataRecord(
            self.schema,
            source_id=self.source_id,
            parent_id=self.id,
            cardinality_idx=self.cardinality_idx,
        )

        # get the set of fields to copy from the parent record
        copy_field_names = project_cols if project_cols is not None else self.get_field_names()
        copy_field_names = [field.split(".")[-1] for field in copy_field_names]

        # copy field types and values from the parent
        for field_name in copy_field_names:
            field_type = self.get_field_type(field_name)
            field_value = self[field_name]
            if (
                not include_bytes
                and isinstance(field_value, bytes)
                or (isinstance(field_value, list) and len(field_value) > 0 and isinstance(field_value[0], bytes))
            ):
                continue

            # set field and value
            new_dr.field_types[field_name] = field_type
            new_dr[field_name] = field_value

        return new_dr


    @staticmethod
    def from_parent(
        schema: Schema,
        parent_record: DataRecord,
        project_cols: list[str] | None = None,
        cardinality_idx: int | None = None,
    ) -> DataRecord:
        # project_cols must be None or contain at least one column
        assert project_cols is None or len(project_cols) > 1, "must have at least one column if using projection"

        # if project_cols is None, then the new schema is a union of the provided schema and parent_record.schema;
        # otherwise, it's a ProjectSchema
        new_schema = schema.union(parent_record.schema)
        if project_cols is not None:
            new_schema = new_schema.project(project_cols)

        # make new record which has parent_record as its parent (and the same source_id)
        new_dr = DataRecord(
            new_schema,
            source_id=parent_record.source_id,
            parent_id=parent_record.id,
            cardinality_idx=cardinality_idx,
        )

        # get the set of fields and field descriptions to copy from the parent record
        copy_field_names = project_cols if project_cols is not None else parent_record.get_field_names()
        copy_field_names = [field.split(".")[-1] for field in copy_field_names]

        # copy fields from the parent
        for field_name in copy_field_names:
            new_dr.field_types[field_name] = parent_record.get_field_type(field_name)
            new_dr[field_name] = parent_record[field_name]

        return new_dr


    @staticmethod
    def from_agg_parents(
        schema: Schema,
        parent_records: DataRecordSet,
        project_cols: list[str] | None = None,
        cardinality_idx: int | None = None,
    ) -> DataRecord:
        # TODO: we can implement this once we support having multiple parent ids
        pass


    @staticmethod
    def from_join_parents(
        left_schema: Schema,
        right_schema: Schema,
        left_parent_record: DataRecord,
        right_parent_record: DataRecord,
        project_cols: list[str] | None = None,
        cardinality_idx: int = None,
    ) -> DataRecord:
        # TODO: we can implement this method if/when we add joins
        pass

    @staticmethod
    def _build_source_id_from_df(source_id: int | str | None = None) -> int | str:
        updated_source_id = source_id
        if source_id is None:
            updated_source_id = "None"
        elif isinstance(source_id, int):
            updated_source_id = str(source_id)
        return f"{FROM_DF_PREFIX}_{updated_source_id}"
    
    @staticmethod
    def _build_schema_from_df(df: pd.DataFrame) -> Schema:
        # Create a unique schema name based on columns
        schema_name = f"{DERIVED_SCHEMA_PREFIX}{hash_for_temp_schema(str(tuple(sorted(df.columns))))}"
        
        if schema_name in globals():
            return globals()[schema_name]
            
        # Create new schema only if it doesn't exist
        new_schema = type(schema_name, (Schema,), {
            '_desc': "Derived schema from DataFrame",
            '__module__': Schema.__module__
        })
        
        for col in df.columns:
            # NOTE: we may need some way of inferring whether fields are images
            setattr(new_schema, col, Field(desc=f"{col}"))
        
        # Store the schema class globally
        globals()[schema_name] = new_schema
        return new_schema
    
    @staticmethod
    def from_df(df: pd.DataFrame, schema: Schema = None, source_id: int | str | None = None) -> list[DataRecord]:
        """Create a list of DataRecords from a pandas DataFrame
        
        Args:
            df (pd.DataFrame): Input DataFrame
            schema (Schema, optional): Schema for the DataRecords. If None, will be derived from DataFrame  
            source_id (int | str | None, optional)
        
        Returns:
            list[DataRecord]: List of DataRecord instances
        """
        if df is None:
            raise ValueError("DataFrame is None!")

        records = []
        if schema is None:
            schema = DataRecord._build_schema_from_df(df)

        field_map = schema.field_map()
        source_id = DataRecord._build_source_id_from_df(source_id)
        for _, row in df.iterrows():
            row_dict = row.to_dict()
            record = DataRecord(schema=schema, source_id=source_id)
            record.field_values = row_dict
            record.field_types = {field_name: field_map[field_name] for field_name in row_dict}
            records.append(record)

        return records
    
    @staticmethod
    def as_df(records: list[DataRecord]) -> pd.DataFrame:
        return pd.DataFrame([record.as_dict() for record in records])

    def as_json_str(self, include_bytes: bool = True, project_cols: list[str] | None = None):
        """Return a JSON representation of this DataRecord"""
        record_dict = self.as_dict(include_bytes, project_cols)
        record_dict = {
            field_name: self.schema.field_to_json(field_name, field_value)
            for field_name, field_value in record_dict.items()
        }
        return json.dumps(record_dict, indent=2)

    def as_dict(self, include_bytes: bool = True, project_cols: list[str] | None = None):
        """Return a dictionary representation of this DataRecord"""
        dct = self.field_values.copy()

        if project_cols is not None and len(project_cols) > 0:
            project_field_names = set(field.split(".")[-1] for field in project_cols)
            dct = {k: v for k, v in dct.items() if k in project_field_names}

        if not include_bytes:
            for k, v in dct.items():
                if isinstance(v, bytes) or (isinstance(v, list) and len(v) > 0 and isinstance(v[0], bytes)):
                    dct[k] = "<bytes>"

        return dct


class DataRecordSet:
    """
    A DataRecordSet contains a list of DataRecords that share the same schema, same parent_id, and same source_id.

    We explicitly check that this is True.

    The record_op_stats could be empty if the DataRecordSet is not from executing an operator.
    """
    def __init__(self, data_records: list[DataRecord], record_op_stats: list[RecordOpStats]):
        # check that all data_records are derived from the same parent record
        if len(data_records) > 0:
            parent_id = data_records[0].parent_id
            error_msg = "DataRecordSet must be constructed from the output of executing a single operator on a single input."
            assert all([dr.parent_id == parent_id for dr in data_records]), error_msg

        # set data_records, parent_id, and source_id; note that it is possible for
        # data_records to be an empty list in the event of a failed convert
        self.data_records = data_records
        self.parent_id = data_records[0].parent_id if len(data_records) > 0 else None
        self.source_id = data_records[0].source_id if len(data_records) > 0 else None

        # set statistics for generating these records
        self.record_op_stats = record_op_stats


    def __getitem__(self, slice):
        return self.data_records[slice]


    def __len__(self):
        return len(self.data_records)


    def __iter__(self):
        yield from self.data_records



import time
from concurrent.futures import ThreadPoolExecutor, wait
from functools import partial
from typing import Callable

import numpy as np

from palimpzest.constants import PARALLEL_EXECUTION_SLEEP_INTERVAL_SECS, OptimizationStrategy
from palimpzest.core.data.dataclasses import ExecutionStats, OperatorStats, PlanStats, RecordOpStats
from palimpzest.core.elements.records import DataRecord, DataRecordSet
from palimpzest.core.lib.schemas import SourceRecord
from palimpzest.policy import Policy
from palimpzest.query.execution.execution_engine import ExecutionEngine
from palimpzest.query.execution.plan_executors.parallel_plan_execution import PipelinedParallelPlanExecutor
from palimpzest.query.execution.plan_executors.single_threaded_plan_execution import SequentialSingleThreadPlanExecutor
from palimpzest.query.operators.convert import ConvertOp, LLMConvert
from palimpzest.query.operators.datasource import CacheScanDataOp, MarshalAndScanDataOp
from palimpzest.query.operators.filter import FilterOp, LLMFilter
from palimpzest.query.operators.physical import PhysicalOperator
from palimpzest.query.operators.retrieve import RetrieveOp
from palimpzest.query.optimizer.cost_model import CostModel, SampleBasedCostModel
from palimpzest.query.optimizer.optimizer import Optimizer
from palimpzest.query.optimizer.plan import SentinelPlan
from palimpzest.sets import Set


class MABSentinelExecutionEngine(ExecutionEngine):
    """
    This class implements the abstract execute() method from the ExecutionEngine.
    This class still needs to be sub-classed by another Execution class which implements
    the higher-level execute_plan() method.
    """
    def __init__(
            self,
            k: int,
            j: int,
            sample_budget: int,
            early_stop_iters: int = 3,
            use_final_op_quality: bool = False,
            seed: int = 42,
            *args,
            **kwargs,
        ):
        super().__init__(*args, **kwargs)
        # self.max_workers = self.get_parallel_max_workers()
        # TODO: undo
        # self.max_workers = 4
        self.k = k
        self.j = j
        self.sample_budget = sample_budget
        self.early_stop_iters = early_stop_iters
        self.use_final_op_quality = use_final_op_quality
        self.pick_output_fn = self.pick_ensemble_output
        self.rng = np.random.default_rng(seed=seed)

    def update_frontier_ops(
        self,
        frontier_ops,
        reservoir_ops,
        policy,
        all_outputs,
        logical_op_id_to_num_samples,
        phys_op_id_to_num_samples,
        is_filter_op_dict,
    ):
        """
        Update the set of frontier operators, pulling in new ones from the reservoir as needed.
        This function will (for each op_set):
        1. Compute the mean, LCB, and UCB for the cost, time, quality, and selectivity of each operator
        2. Compute the pareto optimal set of operators (using the mean values)
        3. Update the frontier and reservoir sets of operators based on their LCB/UCB overlap with the pareto frontier
        """
        # compute metrics for each operator in all_outputs
        logical_op_id_to_op_metrics = {}
        for logical_op_id, source_id_to_record_sets in all_outputs.items():
            # compute selectivity for each physical operator
            phys_op_to_num_inputs, phys_op_to_num_outputs = {}, {}
            for _, record_sets in source_id_to_record_sets.items():
                for record_set in record_sets:
                    op_id = record_set.record_op_stats[0].op_id
                    num_outputs = sum([record_op_stats.passed_operator for record_op_stats in record_set.record_op_stats])
                    if op_id not in phys_op_to_num_inputs:
                        phys_op_to_num_inputs[op_id] = 1
                        phys_op_to_num_outputs[op_id] = num_outputs
                    else:
                        phys_op_to_num_inputs[op_id] += 1
                        phys_op_to_num_outputs[op_id] += num_outputs

            phys_op_to_mean_selectivity = {
                op_id: phys_op_to_num_outputs[op_id] / phys_op_to_num_inputs[op_id]
                for op_id in phys_op_to_num_inputs
            }

            # compute average cost, time, and quality
            phys_op_to_costs, phys_op_to_times, phys_op_to_qualities = {}, {}, {}
            for _, record_sets in source_id_to_record_sets.items():
                for record_set in record_sets:
                    for record_op_stats in record_set.record_op_stats:
                        op_id = record_op_stats.op_id
                        cost = record_op_stats.cost_per_record
                        time = record_op_stats.time_per_record
                        quality = record_op_stats.quality
                        if op_id not in phys_op_to_costs:
                            phys_op_to_costs[op_id] = [cost]
                            phys_op_to_times[op_id] = [time]
                            phys_op_to_qualities[op_id] = [quality]
                        else:
                            phys_op_to_costs[op_id].append(cost)
                            phys_op_to_times[op_id].append(time)
                            phys_op_to_qualities[op_id].append(quality)

            phys_op_to_mean_cost = {op: np.mean(costs) for op, costs in phys_op_to_costs.items()}
            phys_op_to_mean_time = {op: np.mean(times) for op, times in phys_op_to_times.items()}
            phys_op_to_mean_quality = {op: np.mean(qualities) for op, qualities in phys_op_to_qualities.items()}

            # compute average, LCB, and UCB of each operator; the confidence bounds depend upon
            # the computation of the alpha parameter, which we scale to be 0.5 * the mean (of means)
            # of the metric across all operators in this operator set
            cost_alpha = 0.5 * np.mean([mean_cost for mean_cost in phys_op_to_mean_cost.values()])
            time_alpha = 0.5 * np.mean([mean_time for mean_time in phys_op_to_mean_time.values()])
            quality_alpha = 0.5 * np.mean([mean_quality for mean_quality in phys_op_to_mean_quality.values()])
            selectivity_alpha = 0.5 * np.mean([mean_selectivity for mean_selectivity in phys_op_to_mean_selectivity.values()])
 
            op_metrics = {}
            for op_id in phys_op_to_costs:
                sample_ratio = np.sqrt(np.log(logical_op_id_to_num_samples[logical_op_id]) / phys_op_id_to_num_samples[op_id])
                exploration_terms = np.array([cost_alpha * sample_ratio, time_alpha * sample_ratio, quality_alpha * sample_ratio, selectivity_alpha * sample_ratio])
                mean_terms = (phys_op_to_mean_cost[op_id], phys_op_to_mean_time[op_id], phys_op_to_mean_quality[op_id], phys_op_to_mean_selectivity[op_id])

                # NOTE: we could clip these; however I will not do so for now to allow for arbitrary quality metric(s)
                lcb_terms = mean_terms - exploration_terms
                ucb_terms = mean_terms + exploration_terms
                op_metrics[op_id] = {"mean": mean_terms, "lcb": lcb_terms, "ucb": ucb_terms}

            # store average metrics for each operator in the op_set
            logical_op_id_to_op_metrics[logical_op_id] = op_metrics

        # get the tuple representation of this policy
        policy_dict = policy.get_dict()

        # compute the pareto optimal set of operators for each logical_op_id
        pareto_op_sets = {}
        for logical_op_id, op_metrics in logical_op_id_to_op_metrics.items():
            pareto_op_sets[logical_op_id] = set()
            for op_id, metrics in op_metrics.items():
                cost, time, quality, selectivity = metrics["mean"]
                pareto_frontier = True

                # check if any other operator dominates op_id
                for other_op_id, other_metrics in op_metrics.items():
                    other_cost, other_time, other_quality, other_selectivity = other_metrics["mean"]
                    if op_id == other_op_id:
                        continue

                    # if op_id is dominated by other_op_id, set pareto_frontier = False and break
                    # NOTE: here we use a strict inequality (instead of the usual <= or >=) because
                    #       all ops which have equal cost / time / quality / sel. should not be
                    #       filtered out from sampling by our logic in this function
                    cost_dominated = True if policy_dict["cost"] == 0.0 else other_cost < cost
                    time_dominated = True if policy_dict["time"] == 0.0 else other_time < time
                    quality_dominated = True if policy_dict["quality"] == 0.0 else other_quality > quality
                    selectivity_dominated = True if not is_filter_op_dict[logical_op_id] else other_selectivity < selectivity
                    if cost_dominated and time_dominated and quality_dominated and selectivity_dominated:
                        pareto_frontier = False
                        break

                # add op_id to pareto frontier if it's not dominated
                if pareto_frontier:
                    pareto_op_sets[logical_op_id].add(op_id)

        # iterate over frontier ops and replace any which do not overlap with pareto frontier
        new_frontier_ops = {logical_op_id: [] for logical_op_id in frontier_ops}
        new_reservoir_ops = {logical_op_id: [] for logical_op_id in reservoir_ops}
        for logical_op_id, pareto_op_set in pareto_op_sets.items():
            num_dropped_from_frontier = 0
            for op, next_shuffled_sample_idx, new_operator, fully_sampled in frontier_ops[logical_op_id]:
                op_id = op.get_op_id()

                # if this op is fully sampled, remove it from the frontier
                if fully_sampled:
                    num_dropped_from_frontier += 1
                    continue

                # if this op is pareto optimal keep it in our frontier ops
                if op_id in pareto_op_set:
                    new_frontier_ops[logical_op_id].append((op, next_shuffled_sample_idx, new_operator, fully_sampled))
                    continue

                # otherwise, if this op overlaps with an op on the pareto frontier, keep it in our frontier ops
                # NOTE: for now, we perform an optimistic comparison with the ucb/lcb
                pareto_frontier = True
                op_cost = logical_op_id_to_op_metrics[logical_op_id][op_id]["lcb"][0]
                op_time = logical_op_id_to_op_metrics[logical_op_id][op_id]["lcb"][1]
                op_quality = logical_op_id_to_op_metrics[logical_op_id][op_id]["ucb"][2]
                op_selectivity = logical_op_id_to_op_metrics[logical_op_id][op_id]["lcb"][3]
                for pareto_op_id in pareto_op_set:
                    pareto_cost = logical_op_id_to_op_metrics[logical_op_id][pareto_op_id]["ucb"][0]
                    pareto_time = logical_op_id_to_op_metrics[logical_op_id][pareto_op_id]["ucb"][1]
                    pareto_quality = logical_op_id_to_op_metrics[logical_op_id][pareto_op_id]["lcb"][2]
                    pareto_selectivity = logical_op_id_to_op_metrics[logical_op_id][pareto_op_id]["ucb"][3]

                    # if op_id is dominated by pareto_op_id, set pareto_frontier = False and break
                    cost_dominated = True if policy_dict["cost"] == 0.0 else pareto_cost <= op_cost
                    time_dominated = True if policy_dict["time"] == 0.0 else pareto_time <= op_time
                    quality_dominated = True if policy_dict["quality"] == 0.0 else pareto_quality >= op_quality
                    selectivity_dominated = True if not is_filter_op_dict[logical_op_id] else pareto_selectivity <= op_selectivity
                    if cost_dominated and time_dominated and quality_dominated and selectivity_dominated:
                        pareto_frontier = False
                        break
                
                # add op_id to pareto frontier if it's not dominated
                if pareto_frontier:
                    new_frontier_ops[logical_op_id].append((op, next_shuffled_sample_idx, new_operator, fully_sampled))
                else:
                    num_dropped_from_frontier += 1

            # replace the ops dropped from the frontier with new ops from the reservoir
            num_dropped_from_frontier = min(num_dropped_from_frontier, len(reservoir_ops[logical_op_id]))
            for idx in range(num_dropped_from_frontier):
                new_frontier_ops[logical_op_id].append((reservoir_ops[logical_op_id][idx], 0, True, False))

            # update reservoir ops for this logical_op_id
            new_reservoir_ops[logical_op_id] = reservoir_ops[logical_op_id][num_dropped_from_frontier:]

        return new_frontier_ops, new_reservoir_ops


    def compute_quality(
            self,
            record_set: DataRecordSet,
            expected_record_set: DataRecordSet | None = None,
            champion_record_set: DataRecordSet | None = None,
            is_filter_op: bool = False,
            is_convert_op: bool = False,
            field_to_metric_fn: dict[str, str | Callable] | None = None,
        ) -> DataRecordSet:
        """
        Compute the quality for the given `record_set` by comparing it to the `expected_record_set`.

        Update the record_set by assigning the quality to each entry in its record_op_stats and
        returning the updated record_set.
        """
        # compute whether we can only use the champion
        only_using_champion = expected_record_set is None

        # if this operation is a failed convert
        if is_convert_op and len(record_set) == 0:
            record_set.record_op_stats[0].quality = 0.0

        # if this operation is a filter:
        # - we assign a quality of 1.0 if the record is in the expected outputs and it passes this filter
        # - we assign a quality of 0.0 if the record is in the expected outputs and it does NOT pass this filter
        # - we assign a quality relative to the champion / ensemble output if the record is not in the expected outputs
        # we cannot know for certain what the correct behavior is a given filter on a record which is not in the output
        # (unless it is the only filter in the plan), thus we only evaluate the filter based on its performance on
        # records which are in the output
        elif is_filter_op:
            # NOTE:
            # - we know that record_set.record_op_stats will contain a single entry for a filter op
            # - if we are using the champion, then champion_record_set will also contain a single entry for a filter op
            record_op_stats = record_set.record_op_stats[0]
            if only_using_champion:
                champion_record = champion_record_set[0]
                record_op_stats.quality = int(record_op_stats.passed_operator == champion_record.passed_operator)

            # - if we are using validation data, we may have multiple expected records in the expected_record_set for this source_id,
            #   thus, if we can identify an exact match, we can use that to evaluate the filter's quality
            # - if we are using validation data but we *cannot* find an exact match, then we will once again use the champion record set
            else:
                # compute number of matches between this record's computed fields and this expected record's outputs
                found_match_in_output = False
                for expected_record in expected_record_set:
                    all_correct = True
                    for field, value in record_op_stats.record_state.items():
                        if value != getattr(expected_record, field):
                            all_correct = False
                            break

                    if all_correct:
                        found_match_in_output = True
                        break

                if found_match_in_output:
                    record_op_stats.quality = int(record_op_stats.passed_operator == expected_record.passed_operator)
                else:
                    champion_record = champion_record_set[0]
                    record_op_stats.quality = int(record_op_stats.passed_operator == champion_record.passed_operator)

        # if this is a successful convert operation
        else:
            # NOTE: the following computation assumes we do not project out computed values
            #       (and that the validation examples provide all computed fields); even if
            #       a user program does add projection, we can ignore the projection on the
            #       validation dataset and use the champion model (as opposed to the validation
            #       output) for scoring fields which have their values projected out

            # set the expected_record_set to be the champion_record_set if we do not have validation data
            expected_record_set = champion_record_set if only_using_champion else expected_record_set

            # GREEDY ALGORITHM
            # for each record in the expected output, we look for the computed record which maximizes the quality metric;
            # once we've identified that computed record we remove it from consideration for the next expected output
            for expected_record in expected_record_set:
                best_quality, best_record_op_stats = 0.0, None
                for record_op_stats in record_set.record_op_stats:
                    # if we already assigned this record a quality, skip it
                    if record_op_stats.quality is not None:
                        continue

                    # compute number of matches between this record's computed fields and this expected record's outputs
                    total_quality = 0
                    for field in record_op_stats.generated_fields:
                        computed_value = record_op_stats.record_state.get(field, None)
                        expected_value = getattr(expected_record, field)

                        # get the metric function for this field
                        metric_fn = (
                            field_to_metric_fn[field]
                            if field_to_metric_fn is not None and field in field_to_metric_fn
                            else "exact"
                        )

                        # compute exact match
                        if metric_fn == "exact":
                            total_quality += int(computed_value == expected_value)

                        # compute UDF metric
                        elif callable(metric_fn):
                            total_quality += metric_fn(computed_value, expected_value)

                        # otherwise, throw an exception
                        else:
                            raise Exception(f"Unrecognized metric_fn: {metric_fn}")

                    # compute recall and update best seen so far
                    quality = total_quality / len(record_op_stats.generated_fields)
                    if quality > best_quality:
                        best_quality = quality
                        best_record_op_stats = record_op_stats

                # set best_quality as quality for the best_record_op_stats
                if best_record_op_stats is not None:
                    best_record_op_stats.quality = best_quality

        # for any records which did not receive a quality, set it to 0.0 as these are unexpected extras
        for record_op_stats in record_set.record_op_stats:
            if record_op_stats.quality is None:
                record_op_stats.quality = 0.0

        return record_set


    def score_quality(
            self,
            op_set: list[PhysicalOperator],
            logical_op_id: str,
            execution_data: dict[str, dict[str, list[DataRecordSet]]],
            champion_outputs: dict[str, dict[str, DataRecordSet]],
            expected_outputs: dict[str, DataRecordSet] | None = None,
            field_to_metric_fn: dict[str, str | Callable] | None = None,
        ) -> list[RecordOpStats]:
        """
        NOTE: This approach to cost modeling does not work directly for aggregation queries;
              for these queries, we would ask the user to provide validation data for the step immediately
              before a final aggregation

        NOTE: This function currently assumes that one-to-many converts do NOT create duplicate outputs.
        This assumption would break if, for example, we extracted the breed of every dog in an image.
        If there were two golden retrievers and a bernoodle in an image and we extracted:

            {"image": "file1.png", "breed": "Golden Retriever"}
            {"image": "file1.png", "breed": "Golden Retriever"}
            {"image": "file1.png", "breed": "Bernedoodle"}
        
        This function would currently give perfect accuracy to the following output:

            {"image": "file1.png", "breed": "Golden Retriever"}
            {"image": "file1.png", "breed": "Bernedoodle"}

        Even though it is missing one of the golden retrievers.
        """
        # extract information about the logical operation performed at this stage of the sentinel plan;
        # NOTE: we can infer these fields from context clues, but in the long-term we should have a more
        #       principled way of getting these directly from attributes either stored in the sentinel_plan
        #       or in the PhysicalOperator
        physical_op = op_set[0]
        is_filter_op = isinstance(physical_op, FilterOp)
        is_convert_op = isinstance(physical_op, ConvertOp)
        is_perfect_quality_op = (
            not isinstance(physical_op, LLMConvert)
            and not isinstance(physical_op, LLMFilter)
            and not isinstance(physical_op, RetrieveOp)
        )

        # pull out the execution data from this operator; place the upstream execution data in a new list
        this_op_execution_data = execution_data[logical_op_id]

        # compute quality of each output computed by this operator
        for source_id, record_sets in this_op_execution_data.items():
            # NOTE
            # source_id is a particular input, for which we may have computed multiple output record_sets;
            # each of these record_sets may contain more than one record (b/c one-to-many) and we have one
            # record_set per operator in the op_set

            # if this operation does not involve an LLM, every record_op_stats object gets perfect quality
            if is_perfect_quality_op:
                for record_set in record_sets:
                    for record_op_stats in record_set.record_op_stats:
                        record_op_stats.quality = 1.0
                continue

            # get the expected output for this source_id if we have one
            expected_record_set = (
                expected_outputs[source_id]
                if expected_outputs is not None and source_id in expected_outputs
                else None
            )

            # extract champion output for this record set
            champion_record_set = champion_outputs[logical_op_id][source_id]

            # for each record_set produced by an operation, compute its quality
            for record_set in record_sets:
                record_set = self.compute_quality(record_set, expected_record_set, champion_record_set, is_filter_op, is_convert_op, field_to_metric_fn)

        # return the quality annotated record op stats
        return execution_data


    def pick_ensemble_output(self, op_set_record_sets: list[tuple[DataRecordSet, PhysicalOperator]]) -> DataRecordSet:
        # if there's only one operator in the set, we return its record_set
        if len(op_set_record_sets) == 1:
            record_set, _ = op_set_record_sets[0]
            return record_set

        # NOTE: I don't like that this assumes the models are consistent in
        #       how they order their record outputs for one-to-many converts;
        #       eventually we can try out more robust schemes to account for
        #       differences in ordering
        # aggregate records at each index in the response
        idx_to_records = {}
        for record_set, _ in op_set_record_sets:
            for idx, record in enumerate(record_set):
                if idx not in idx_to_records:
                    idx_to_records[idx] = [record]
                else:
                    idx_to_records[idx].append(record)

        # compute most common answer at each index
        out_records = []
        for idx in range(len(idx_to_records)):
            records = idx_to_records[idx]
            most_common_record = max(set(records), key=records.count)
            out_records.append(most_common_record)

        # create and return final DataRecordSet
        return DataRecordSet(out_records, [])


    def pick_highest_quality_output(self, op_set_record_sets: list[tuple[DataRecordSet, PhysicalOperator]]) -> DataRecordSet:
        # if there's only one operator in the set, we return its record_set
        if len(op_set_record_sets) == 1:
            record_set, _ = op_set_record_sets[0]
            return record_set

        # aggregate records at each index in the response
        idx_to_records = {}
        for record_set, _ in op_set_record_sets:
            for idx in range(len(record_set)):
                record, record_op_stats = record_set[idx], record_set.record_op_stats[idx]
                if idx not in idx_to_records:
                    idx_to_records[idx] = [(record, record_op_stats)]
                else:
                    idx_to_records[idx].append((record, record_op_stats))

        # compute highest quality answer at each index
        out_records = []
        for idx in range(len(idx_to_records)):
            records_lst, record_op_stats_lst = zip(*idx_to_records[idx])
            max_quality_record, max_quality = records_lst[0], record_op_stats_lst[0].quality
            for record, record_op_stats in zip(records_lst[1:], record_op_stats_lst[1:]):
                record_quality = record_op_stats.quality
                if record_quality > max_quality:
                    max_quality_record = record
                    max_quality = record_quality
            out_records.append(max_quality_record)

        # create and return final DataRecordSet
        return DataRecordSet(out_records, [])


    @staticmethod
    def execute_op_wrapper(operator: PhysicalOperator, op_input: DataRecord | list[DataRecord]) -> tuple[DataRecordSet, PhysicalOperator]:
        """
        Wrapper function around operator execution which also and returns the operator.
        This is useful in the parallel setting(s) where operators are executed by a worker pool,
        and it is convenient to return the op_id along with the computation result.
        """
        record_set = operator(op_input)

        return record_set, operator, op_input


    def execute_op_set(self, op_candidate_pairs):
        # TODO: post-submission we will need to modify this to:
        # - submit all candidates for aggregate operators
        # - handle limits
        # create thread pool w/max workers and run futures over worker pool
        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            # create futures
            futures = []
            for operator, candidate in op_candidate_pairs:
                future = executor.submit(MABSentinelExecutionEngine.execute_op_wrapper, operator, candidate)
                futures.append(future)

            # compute output record_set for each (operator, candidate) pair
            output_record_sets = []
            while len(futures) > 0:
                # get the set of futures that have (and have not) finished in the last PARALLEL_EXECUTION_SLEEP_INTERVAL_SECS
                done_futures, not_done_futures = wait(futures, timeout=PARALLEL_EXECUTION_SLEEP_INTERVAL_SECS)

                # cast not_done_futures from a set to a list so we can append to it
                not_done_futures = list(not_done_futures)

                # process finished futures
                for future in done_futures:
                    # get the result and add it to the output records set
                    record_set, operator, candidate = future.result()
                    output_record_sets.append((record_set, operator, candidate))

                # update list of futures
                futures = not_done_futures

            # compute mapping from source_id to record sets for all operators and for champion operator
            all_record_sets, champion_record_sets = {}, {}
            for op, candidate in op_candidate_pairs:
                candidate_output_record_sets, source_id = [], None
                for record_set, operator, candidate_ in output_record_sets:
                    if candidate == candidate_:
                        candidate_output_record_sets.append((record_set, operator))

                        # NOTE: we should resolve this issue in a more thoughtful way, but currently the source_id for
                        #       scan candidate records is the sample_idx, when we want it to be the source_id which is
                        #       set by the scan operator when it returns the record
                        # get the source_id associated with this input record
                        source_id = (
                            candidate.source_id
                            if not (isinstance(op, (MarshalAndScanDataOp, CacheScanDataOp)))
                            else record_set[0].source_id
                        )

                # select the champion (i.e. best) record_set from all the record sets computed for this candidate
                champion_record_set = self.pick_output_fn(candidate_output_record_sets)

                # add champion record_set to mapping from source_id --> champion record_set
                champion_record_sets[source_id] = champion_record_set

                # add all record_sets computed for this source_id to mapping from source_id --> record_sets
                all_record_sets[source_id] = [tup[0] for tup in candidate_output_record_sets]

        return all_record_sets, champion_record_sets


    def execute_sentinel_plan(self, plan: SentinelPlan, expected_outputs: dict[str, DataRecordSet], policy: Policy):
        """
        """
        if self.verbose:
            print("----------------------")
            print(f"PLAN[{plan.plan_id}] (sentinel):")
            print(plan)
            print("---")

        plan_start_time = time.time()

        # initialize plan stats and operator stats
        plan_stats = PlanStats(plan_id=plan.plan_id, plan_str=str(plan))
        for logical_op_id, logical_op_name, op_set in plan:
            op_set_details = {
                op.op_name(): {k: str(v) for k, v in op.get_id_params().items()}
                for op in op_set
            }
            plan_stats.operator_stats[logical_op_id] = OperatorStats(
                op_id=logical_op_id,
                op_name=logical_op_name,
                op_details=op_set_details,
            )

        # shuffle the indices of records to sample
        total_num_samples = self.datasource.get_val_length()
        shuffled_sample_indices = [int(idx) for idx in np.arange(total_num_samples)]
        self.rng.shuffle(shuffled_sample_indices)

        # sample k initial operators for each operator set; for each operator maintain a tuple of:
        # (operator, next_shuffled_sample_idx, new_operator); new_operator is True when an operator
        # is added to the frontier
        frontier_ops, reservoir_ops = {}, {}
        for logical_op_id, _, op_set in plan:
            op_set_copy = [op for op in op_set]
            self.rng.shuffle(op_set_copy)
            k = min(self.k, len(op_set_copy))
            frontier_ops[logical_op_id] = [(op, 0, True, False) for op in op_set_copy[:k]]
            reservoir_ops[logical_op_id] = [op for op in op_set_copy[k:]]

        # create mapping from logical and physical op ids to the number of samples drawn
        logical_op_id_to_num_samples = {logical_op_id: 0 for logical_op_id, _, _ in plan}
        phys_op_id_to_num_samples = {op.get_op_id(): 0 for _, _, op_set in plan for op in op_set}
        is_filter_op_dict = {
            logical_op_id: isinstance(op_set[0], FilterOp)
            for logical_op_id, _, op_set in plan
        }

        # TODO: long-term, we should do something which does not rely on scanning validation source to build this mapping
        sample_idx_to_source_id = {
            sample_idx: self.datasource.get_item(sample_idx, val=True).source_id
            for sample_idx in range(total_num_samples)
        }

        # NOTE: to maintain parity with our count of samples drawn in the random sampling execution,
        # for each logical_op_id, we count the number of (record, op) executions as the number of samples within that op_set;
        # the samples drawn is equal to the max of that number across all operator sets
        samples_drawn = 0
        all_outputs, champion_outputs = {}, {}
        while samples_drawn < self.sample_budget:
            # execute operator sets in sequence
            for op_idx, (logical_op_id, _, op_set) in enumerate(plan):
                prev_logical_op_id = plan.logical_op_ids[op_idx - 1] if op_idx > 0 else None
                prev_logical_op_is_filter =  prev_logical_op_id is not None and is_filter_op_dict[prev_logical_op_id]

                # create list of tuples for (op, candidate) which we should execute
                op_candidate_pairs = []
                updated_frontier_ops_lst = []
                for op, next_shuffled_sample_idx, new_operator, fully_sampled in frontier_ops[logical_op_id]:
                    # execute new operators on first j candidates, and previously sampled operators on one additional candidate
                    j = min(self.j, len(shuffled_sample_indices)) if new_operator else 1
                    for j_idx in range(j):
                        candidates = []
                        if isinstance(op, (MarshalAndScanDataOp, CacheScanDataOp)):
                            sample_idx = shuffled_sample_indices[(next_shuffled_sample_idx + j_idx) % len(shuffled_sample_indices)]
                            candidate = DataRecord(schema=SourceRecord, source_id=sample_idx)
                            candidate.idx = sample_idx
                            candidate.get_item_fn = partial(self.datasource.get_item, val=True)
                            candidates = [candidate]
                            logical_op_id_to_num_samples[logical_op_id] += 1
                            phys_op_id_to_num_samples[op.get_op_id()] += 1
                        else:
                            if next_shuffled_sample_idx + j_idx == len(shuffled_sample_indices):
                                fully_sampled = True
                                break

                            # pick best output from all_outputs from previous logical operator
                            sample_idx = shuffled_sample_indices[next_shuffled_sample_idx + j_idx]
                            source_id = sample_idx_to_source_id[sample_idx]
                            record_sets = all_outputs[prev_logical_op_id][source_id]
                            all_source_record_sets = [(record_set, None) for record_set in record_sets]
                            max_quality_record_set = self.pick_highest_quality_output(all_source_record_sets)
                            if (
                                not prev_logical_op_is_filter
                                or (
                                    prev_logical_op_is_filter
                                    and max_quality_record_set.record_op_stats[0].passed_operator
                                )
                            ):
                                candidates = [record for record in max_quality_record_set]

                            # increment number of samples drawn for this logical and physical op id; even if we get multiple
                            # candidates from the previous stage in the pipeline, we only count this as one sample
                            logical_op_id_to_num_samples[logical_op_id] += 1
                            phys_op_id_to_num_samples[op.get_op_id()] += 1

                        if len(candidates) > 0:
                            op_candidate_pairs.extend([(op, candidate) for candidate in candidates])

                    # set new_operator = False and update next_shuffled_sample_idx
                    updated_frontier_ops_lst.append((op, next_shuffled_sample_idx + j, False, fully_sampled))

                frontier_ops[logical_op_id] = updated_frontier_ops_lst

                # continue if op_candidate_pairs is an empty list, as this means all records have been filtered out
                if len(op_candidate_pairs) == 0:
                    continue

                # run sampled operators on sampled candidates
                source_id_to_record_sets, source_id_to_champion_record_set = self.execute_op_set(op_candidate_pairs)

                # update all_outputs and champion_outputs dictionary
                if logical_op_id not in all_outputs:
                    all_outputs[logical_op_id] = source_id_to_record_sets
                    champion_outputs[logical_op_id] = source_id_to_champion_record_set
                else:
                    for source_id, record_sets in source_id_to_record_sets.items():
                        if source_id not in all_outputs[logical_op_id]:
                            all_outputs[logical_op_id][source_id] = record_sets
                            champion_outputs[logical_op_id][source_id] = source_id_to_champion_record_set[source_id]
                        else:
                            all_outputs[logical_op_id][source_id].extend(record_sets)
                            # NOTE: short-term solution; in practice we can get multiple champion records from different
                            # sets of operators, so we should try to find a way to only take one
                            champion_outputs[logical_op_id][source_id] = source_id_to_champion_record_set[source_id]

                # flatten lists of records and record_op_stats
                all_records, all_record_op_stats = [], []
                for _, record_sets in source_id_to_record_sets.items():
                    for record_set in record_sets:
                        all_records.extend(record_set.data_records)
                        all_record_op_stats.extend(record_set.record_op_stats)

                # update plan stats
                plan_stats.operator_stats[logical_op_id].add_record_op_stats(
                    all_record_op_stats,
                    source_op_id=prev_logical_op_id,
                    plan_id=plan.plan_id,
                )

                # add records (which are not filtered) to the cache, if allowed
                if not self.nocache:
                    for record in all_records:
                        if getattr(record, "passed_operator", True):
                            self.datadir.append_cache(logical_op_id, record)

                # compute quality for each operator
                field_to_metric_fn = self.datasource.get_field_to_metric_fn()
                all_outputs = self.score_quality(
                    op_set,
                    logical_op_id,
                    all_outputs,
                    champion_outputs,
                    expected_outputs,
                    field_to_metric_fn,
                )

            # update the (pareto) frontier for each set of operators
            frontier_ops, reservoir_ops = self.update_frontier_ops(
                frontier_ops,
                reservoir_ops,
                policy,
                all_outputs,
                logical_op_id_to_num_samples,
                phys_op_id_to_num_samples,
                is_filter_op_dict,
            )

            # update the number of samples drawn to be the max across all logical operators
            samples_drawn = max(logical_op_id_to_num_samples.values())

        # if caching was allowed, close the cache
        if not self.nocache:
            for logical_op_id, _, _ in plan:
                self.datadir.close_cache(logical_op_id)

        # finalize plan stats
        total_plan_time = time.time() - plan_start_time
        plan_stats.finalize(total_plan_time)

        return all_outputs, plan_stats


    def generate_sample_observations(self, sentinel_plan: SentinelPlan, policy: Policy):
        """
        This function is responsible for generating sample observation data which can be
        consumed by the CostModel.

        To accomplish this, we construct a special sentinel plan using the Optimizer which is
        capable of executing any valid physical implementation of a Filter or Convert operator
        on each record.
        """
        # if we're using validation data, get the set of expected output records
        expected_outputs = {}
        for idx in range(self.datasource.get_val_length()):
            data_records = self.datasource.get_item(idx, val=True, include_label=True)
            if not isinstance(data_records, list):
                data_records = [data_records]
            record_set = DataRecordSet(data_records, None)
            expected_outputs[record_set.source_id] = record_set

        # run sentinel plan
        execution_data, plan_stats = self.execute_sentinel_plan(sentinel_plan, expected_outputs, policy)

        return execution_data, plan_stats


    def create_sentinel_plan(self, dataset: Set, policy: Policy) -> SentinelPlan:
        """
        Generates and returns a SentinelPlan for the given dataset.
        """
        # TODO: explicitly pull up filters; for SIGMOD we can explicitly write plans w/filters pulled up
        # initialize the optimizer
        optimizer = Optimizer(
            policy=policy,
            cost_model=CostModel(),
            no_cache=True,
            verbose=self.verbose,
            available_models=self.available_models,
            allow_bonded_query=self.allow_bonded_query,
            allow_conventional_query=self.allow_conventional_query,
            allow_code_synth=self.allow_code_synth,
            allow_token_reduction=self.allow_token_reduction,
            allow_rag_reduction=self.allow_rag_reduction,
            allow_mixtures=self.allow_mixtures,
            optimization_strategy=OptimizationStrategy.SENTINEL,
        )

        # use optimizer to generate sentinel plans
        sentinel_plans = optimizer.optimize(dataset, policy)
        sentinel_plan = sentinel_plans[0]

        return sentinel_plan


    def execute(self, dataset: Set, policy: Policy):
        execution_start_time = time.time()

        # for now, enforce that we are using validation data; we can relax this after paper submission
        if not self.using_validation_data:
            raise Exception("Make sure you are using ValidationDataSource with MABSentinelExecutionEngine")

        # if nocache is True, make sure we do not re-use codegen examples
        if self.nocache:
            self.clear_cached_examples()

        # create sentinel plan
        sentinel_plan = self.create_sentinel_plan(dataset, policy)

        # generate sample execution data
        all_execution_data, plan_stats = self.generate_sample_observations(sentinel_plan, policy)

        # put sentinel plan execution stats into list and prepare list of output records
        all_plan_stats = [plan_stats]
        all_records = []

        # construct the CostModel with any sample execution data we've gathered
        cost_model = SampleBasedCostModel(sentinel_plan, all_execution_data, self.verbose)

        # (re-)initialize the optimizer
        optimizer = Optimizer(
            policy=policy,
            cost_model=cost_model,
            no_cache=self.nocache,
            verbose=self.verbose,
            available_models=self.available_models,
            allow_bonded_query=self.allow_bonded_query,
            allow_conventional_query=self.allow_conventional_query,
            allow_code_synth=self.allow_code_synth,
            allow_token_reduction=self.allow_token_reduction,
            allow_rag_reduction=self.allow_rag_reduction,
            allow_mixtures=self.allow_mixtures,
            optimization_strategy=self.optimization_strategy,
            use_final_op_quality=self.use_final_op_quality,
        )
        total_optimization_time = time.time() - execution_start_time

        # execute plan(s) according to the optimization strategy
        if self.optimization_strategy == OptimizationStrategy.CONFIDENCE_INTERVAL:
            records, plan_stats = self.execute_confidence_interval_strategy(dataset, policy, optimizer)
            all_records.extend(records)
            all_plan_stats.extend(plan_stats)

        else:
            records, plan_stats = self.execute_strategy(dataset, policy, optimizer)
            all_records.extend(records)
            all_plan_stats.extend(plan_stats)

        # aggregate plan stats
        aggregate_plan_stats = self.aggregate_plan_stats(all_plan_stats)

        # add sentinel records and plan stats (if captured) to plan execution data
        execution_stats = ExecutionStats(
            execution_id=self.execution_id(),
            plan_stats=aggregate_plan_stats,
            total_optimization_time=total_optimization_time,
            total_execution_time=time.time() - execution_start_time,
            total_execution_cost=sum(list(map(lambda plan_stats: plan_stats.total_plan_cost, aggregate_plan_stats.values()))),
            plan_strs={plan_id: plan_stats.plan_str for plan_id, plan_stats in aggregate_plan_stats.items()},
        )

        return all_records, execution_stats


class MABSequentialSingleThreadSentinelExecution(MABSentinelExecutionEngine, SequentialSingleThreadPlanExecutor):
    """
    This class performs sentinel execution while executing plans in a sequential, single-threaded fashion.
    """
    def __init__(self, *args, **kwargs):
        MABSentinelExecutionEngine.__init__(self, *args, **kwargs)
        SequentialSingleThreadPlanExecutor.__init__(self, *args, **kwargs)


class MABSequentialParallelSentinelExecution(MABSentinelExecutionEngine, SequentialSingleThreadPlanExecutor):
    """
    This class performs sentinel execution while executing plans in a pipelined, parallel fashion.
    """
    def __init__(self, *args, **kwargs):
        MABSentinelExecutionEngine.__init__(self, *args, **kwargs)
        PipelinedParallelPlanExecutor.__init__(self, *args, **kwargs)



import time

from palimpzest.core.data.dataclasses import OperatorStats, PlanStats
from palimpzest.core.elements.records import DataRecord
from palimpzest.core.lib.schemas import SourceRecord
from palimpzest.query.execution.execution_engine import ExecutionEngine
from palimpzest.query.operators.aggregate import AggregateOp
from palimpzest.query.operators.datasource import DataSourcePhysicalOp, MarshalAndScanDataOp
from palimpzest.query.operators.filter import FilterOp
from palimpzest.query.operators.limit import LimitScanOp
from palimpzest.query.optimizer.plan import PhysicalPlan


class SequentialSingleThreadPlanExecutor(ExecutionEngine):
    """
    This class implements the abstract execute_plan() method from the ExecutionEngine.
    This class still needs to be sub-classed by another Execution class which implements
    the higher-level execute() method.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def execute_plan(self, plan: PhysicalPlan, num_samples: int | float = float("inf"), plan_workers: int = 1):
        """Initialize the stats and the execute the plan."""
        if self.verbose:
            print("----------------------")
            print(f"PLAN[{plan.plan_id}] (n={num_samples}):")
            print(plan)
            print("---")

        plan_start_time = time.time()

        # initialize plan stats and operator stats
        plan_stats = PlanStats(plan_id=plan.plan_id, plan_str=str(plan))
        for op in plan.operators:
            op_id = op.get_op_id()
            op_name = op.op_name()
            op_details = {k: str(v) for k, v in op.get_id_params().items()}
            plan_stats.operator_stats[op_id] = OperatorStats(op_id=op_id, op_name=op_name, op_details=op_details)

        # initialize list of output records and intermediate variables
        output_records = []
        current_scan_idx = self.scan_start_idx

        # get handle to DataSource and pre-compute its size
        source_operator = plan.operators[0]
        datasource = (
            self.datadir.get_registered_dataset(source_operator.dataset_id)
            if isinstance(source_operator, MarshalAndScanDataOp)
            else self.datadir.get_cached_result(source_operator.dataset_id)
        )
        datasource_len = len(datasource)

        # initialize processing queues for each operation
        processing_queues = {op.get_op_id(): [] for op in plan.operators if not isinstance(op, DataSourcePhysicalOp)}

        # execute the plan one operator at a time
        for op_idx, operator in enumerate(plan.operators):
            op_id = operator.get_op_id()
            prev_op_id = plan.operators[op_idx - 1].get_op_id() if op_idx > 1 else None
            next_op_id = plan.operators[op_idx + 1].get_op_id() if op_idx + 1 < len(plan.operators) else None

            # initialize output records and record_op_stats for this operator
            records, record_op_stats = [], []

            # invoke datasource operator(s) until we run out of source records or hit the num_samples limit
            if isinstance(operator, DataSourcePhysicalOp):
                keep_scanning_source_records = True
                while keep_scanning_source_records:
                    # construct input DataRecord for DataSourcePhysicalOp
                    # NOTE: this DataRecord will be discarded and replaced by the scan_operator;
                    #       it is simply a vessel to inform the scan_operator which record to fetch
                    candidate = DataRecord(schema=SourceRecord, source_id=current_scan_idx)
                    candidate.idx = current_scan_idx
                    candidate.get_item_fn = datasource.get_item

                    # run DataSourcePhysicalOp on record
                    record_set = operator(candidate)
                    records.extend(record_set.data_records)
                    record_op_stats.extend(record_set.record_op_stats)

                    # update the current scan index
                    current_scan_idx += 1

                    # update whether to keep scanning source records
                    keep_scanning_source_records = current_scan_idx < datasource_len and len(records) < num_samples

            # aggregate operators accept all input records at once
            elif isinstance(operator, AggregateOp):
                record_set = operator(candidates=processing_queues[op_id])
                records = record_set.data_records
                record_op_stats = record_set.record_op_stats

            # otherwise, process the records in the processing queue for this operator one at a time
            elif len(processing_queues[op_id]) > 0:
                for input_record in processing_queues[op_id]:
                    record_set = operator(input_record)
                    records.extend(record_set.data_records)
                    record_op_stats.extend(record_set.record_op_stats)

                    if isinstance(operator, LimitScanOp) and len(records) == operator.limit:
                        break

            # update plan stats
            plan_stats.operator_stats[op_id].add_record_op_stats(
                record_op_stats,
                source_op_id=prev_op_id,
                plan_id=plan.plan_id,
            )

            # add records (which are not filtered) to the cache, if allowed
            if not self.nocache:
                for record in records:
                    if getattr(record, "passed_operator", True):
                        self.datadir.append_cache(operator.target_cache_id, record)

            # update processing_queues or output_records
            for record in records:
                if isinstance(operator, FilterOp) and not record.passed_operator:
                    continue
                if next_op_id is not None:
                    processing_queues[next_op_id].append(record)
                else:
                    output_records.append(record)

            # if we've filtered out all records, terminate early
            if next_op_id is not None and processing_queues[next_op_id] == []:
                break

        # if caching was allowed, close the cache
        if not self.nocache:
            for operator in plan.operators:
                self.datadir.close_cache(operator.target_cache_id)

        # finalize plan stats
        total_plan_time = time.time() - plan_start_time
        plan_stats.finalize(total_plan_time)

        return output_records, plan_stats


class PipelinedSingleThreadPlanExecutor(ExecutionEngine):
    """
    This class implements the abstract execute_plan() method from the ExecutionEngine.
    This class still needs to be sub-classed by another Execution class which implements
    the higher-level execute() method.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.max_workers = 1 if self.max_workers is None else self.max_workers

    def execute_plan(self, plan: PhysicalPlan, num_samples: int | float = float("inf"), plan_workers: int = 1):
        """Initialize the stats and the execute the plan."""
        if self.verbose:
            print("----------------------")
            print(f"PLAN[{plan.plan_id}] (n={num_samples}):")
            print(plan)
            print("---")

        plan_start_time = time.time()

        # initialize plan stats and operator stats
        plan_stats = PlanStats(plan_id=plan.plan_id, plan_str=str(plan))
        for op in plan.operators:
            op_id = op.get_op_id()
            op_name = op.op_name()
            op_details = {k: str(v) for k, v in op.get_id_params().items()}
            plan_stats.operator_stats[op_id] = OperatorStats(op_id=op_id, op_name=op_name, op_details=op_details)

        # initialize list of output records and intermediate variables
        output_records = []
        source_records_scanned = 0
        current_scan_idx = self.scan_start_idx

        # get handle to DataSource and pre-compute its size
        source_operator = plan.operators[0]
        datasource = (
            self.datadir.get_registered_dataset(source_operator.dataset_id)
            if isinstance(source_operator, MarshalAndScanDataOp)
            else self.datadir.get_cached_result(source_operator.dataset_id)
        )
        datasource_len = len(datasource)

        # initialize processing queues for each operation
        processing_queues = {op.get_op_id(): [] for op in plan.operators if not isinstance(op, DataSourcePhysicalOp)}

        # execute the plan until either:
        # 1. all records have been processed, or
        # 2. the final limit operation has completed
        finished_executing, keep_scanning_source_records = False, True
        while not finished_executing:
            for op_idx, operator in enumerate(plan.operators):
                op_id = operator.get_op_id()

                prev_op_id = plan.operators[op_idx - 1].get_op_id() if op_idx > 1 else None
                next_op_id = plan.operators[op_idx + 1].get_op_id() if op_idx + 1 < len(plan.operators) else None

                # create empty lists for records and execution stats generated by executing this operator on its next input(s)
                records, record_op_stats = [], []

                # invoke datasource operator(s) until we run out of source records or hit the num_samples limit
                if isinstance(operator, DataSourcePhysicalOp):
                    if keep_scanning_source_records:
                        # construct input DataRecord for DataSourcePhysicalOp
                        # NOTE: this DataRecord will be discarded and replaced by the scan_operator;
                        #       it is simply a vessel to inform the scan_operator which record to fetch
                        candidate = DataRecord(schema=SourceRecord, source_id=current_scan_idx)
                        candidate.idx = current_scan_idx
                        candidate.get_item_fn = datasource.get_item

                        # run DataSourcePhysicalOp on record
                        record_set = operator(candidate)
                        records = record_set.data_records
                        record_op_stats = record_set.record_op_stats

                        # update number of source records scanned and the current index
                        source_records_scanned += len(records)
                        current_scan_idx += 1
                    else:
                        continue

                # only invoke aggregate operator(s) once there are no more source records and all
                # upstream operators' processing queues are empty
                elif isinstance(operator, AggregateOp):
                    upstream_ops_are_finished = True
                    for upstream_op_idx in range(op_idx):
                        # datasources do not have processing queues
                        if isinstance(plan.operators[upstream_op_idx], DataSourcePhysicalOp):
                            continue

                        # check upstream ops which do have a processing queue
                        upstream_op_id = plan.operators[upstream_op_idx].get_op_id()
                        upstream_ops_are_finished = (
                            upstream_ops_are_finished and len(processing_queues[upstream_op_id]) == 0
                        )

                    if not keep_scanning_source_records and upstream_ops_are_finished:
                        record_set = operator(candidates=processing_queues[op_id])
                        records = record_set.data_records
                        record_op_stats = record_set.record_op_stats
                        processing_queues[op_id] = []

                # otherwise, process the next record in the processing queue for this operator
                elif len(processing_queues[op_id]) > 0:
                    input_record = processing_queues[op_id].pop(0)
                    record_set = operator(input_record)
                    records = record_set.data_records
                    record_op_stats = record_set.record_op_stats

                # if records were generated by this operator, process them
                if len(records) > 0:
                    # update plan stats
                    plan_stats.operator_stats[op_id].add_record_op_stats(
                        record_op_stats,
                        source_op_id=prev_op_id,
                        plan_id=plan.plan_id,
                    )

                    # add records (which are not filtered) to the cache, if allowed
                    if not self.nocache:
                        for record in records:
                            if getattr(record, "passed_operator", True):
                                self.datadir.append_cache(operator.target_cache_id, record)

                    # update processing_queues or output_records
                    for record in records:
                        if isinstance(operator, FilterOp) and not record.passed_operator:
                            continue
                        if next_op_id is not None:
                            processing_queues[next_op_id].append(record)
                        else:
                            output_records.append(record)

            # update finished_executing based on whether all records have been processed
            still_processing = any([len(queue) > 0 for queue in processing_queues.values()])
            keep_scanning_source_records = current_scan_idx < datasource_len and source_records_scanned < num_samples
            finished_executing = not keep_scanning_source_records and not still_processing

            # update finished_executing based on limit
            if isinstance(operator, LimitScanOp):
                finished_executing = len(output_records) == operator.limit

        # if caching was allowed, close the cache
        if not self.nocache:
            for operator in plan.operators:
                self.datadir.close_cache(operator.target_cache_id)

        # finalize plan stats
        total_plan_time = time.time() - plan_start_time
        plan_stats.finalize(total_plan_time)

        return output_records, plan_stats



from palimpzest.constants import Model, OptimizationStrategy
from palimpzest.core.data.datasources import DataSource
from palimpzest.datamanager.datamanager import DataDirectory
from palimpzest.policy import Policy
from palimpzest.query.execution.execution_engine import ExecutionEngine
from palimpzest.query.execution.nosentinel_execution import NoSentinelSequentialSingleThreadExecution
from palimpzest.sets import Set


class Execute:
    @classmethod
    def get_datasource(cls, dataset: Set | DataSource) -> str:
        """
        Gets the DataSource for the given dataset.
        """
        # iterate until we reach DataSource
        while isinstance(dataset, Set):
            dataset = dataset._source

        # this will throw an exception if datasource is not registered with PZ
        return DataDirectory().get_registered_dataset(dataset.dataset_id)

    def __new__(
        cls,
        dataset: Set,
        policy: Policy,
        num_samples: int = 20,
        nocache: bool = True,
        include_baselines: bool = False,
        min_plans: int | None = None,
        max_workers: int = 1,
        verbose: bool = False,
        available_models: list[Model] | None = None,
        allow_bonded_query: bool = True,
        allow_conventional_query: bool = False,
        allow_model_selection: bool = True,
        allow_code_synth: bool = False,
        allow_token_reduction: bool = False,
        allow_rag_reduction: bool = True,
        allow_mixtures: bool = True,
        optimization_strategy: OptimizationStrategy = OptimizationStrategy.PARETO,
        execution_engine: ExecutionEngine = NoSentinelSequentialSingleThreadExecution,
        *args,
        **kwargs,
    ):
        if available_models is None:
            available_models = []
        return execution_engine(
            *args,
            **kwargs,
            datasource=cls.get_datasource(dataset),
            num_samples=num_samples,
            nocache=nocache,
            include_baselines=include_baselines,
            min_plans=min_plans,
            max_workers=max_workers,
            verbose=verbose,
            available_models=available_models,
            allow_bonded_query=allow_bonded_query,
            allow_conventional_query=allow_conventional_query,
            allow_code_synth=allow_code_synth,
            allow_model_selection=allow_model_selection,
            allow_token_reduction=allow_token_reduction,
            allow_rag_reduction=allow_rag_reduction,
            allow_mixtures=allow_mixtures,
            optimization_strategy=optimization_strategy,
        ).execute(dataset=dataset, policy=policy)



from __future__ import annotations

import os
import time

from palimpzest.core.data.dataclasses import OperatorCostEstimates, RecordOpStats
from palimpzest.core.elements.records import DataRecord, DataRecordSet
from palimpzest.query.operators.physical import PhysicalOperator


class RetrieveOp(PhysicalOperator):
    def __init__(self, index, search_attr, output_attr, k, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.index = index
        self.search_attr = search_attr
        self.output_attr = output_attr
        self.k = k

    def __str__(self):
        op = super().__str__()
        op += f"    Retrieve: {str(self.index)} with top {self.k}\n"
        return op

    def get_id_params(self):
        id_params = super().get_id_params()
        id_params = {
            "search_attr": self.search_attr,
            "output_attr": self.output_attr,
            "k": self.k,
            **id_params,
        }

        return id_params

    def get_op_params(self):
        op_params = super().get_op_params()
        op_params = {
            "index": self.index,
            "search_attr": self.search_attr,
            "output_attr": self.output_attr,
            "k": self.k,
            **op_params,
        }

        return op_params

    def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates) -> OperatorCostEstimates:
        """
        Compute naive cost estimates for the Retrieve operation. These estimates assume
        that the Retrieve (1) has no cost and (2) has perfect quality.
        """
        return OperatorCostEstimates(
            cardinality=source_op_cost_estimates.cardinality,
            time_per_record=0.001,  # estimate 1 ms single-threaded execution for index lookup
            cost_per_record=0.0,
            quality=1.0,
        )

    def __call__(self, candidate: DataRecord) -> DataRecordSet:
        start_time = time.time()

        query = getattr(candidate, self.search_attr)

        top_k_results, top_k_result_doc_ids = [], []
        if isinstance(query, str):
            results = self.index.search(query, k=self.k)
            top_k_results = [result["content"] for result in results]

            # This is hacky, fix this later.
            top_k_result_doc_ids = list({result["document_id"] for result in results})

        elif isinstance(query, list):
            try:
                # retrieve top entry for each query
                results = self.index.search(query, k=1)

                # filter for the top-k entries
                results = [result[0] if isinstance(result, list) else result for result in results]
                sorted_results = sorted(results, key=lambda result: result["score"], reverse=True)
                top_k_results = [result["content"] for result in sorted_results[:self.k]]
                top_k_result_doc_ids = [result["document_id"] for result in sorted_results[:self.k]]
            except Exception:
                os.makedirs("retrieve-errors", exist_ok=True)
                ts = time.time()
                with open(f"retrieve-errors/error-{ts}.txt", "w") as f:
                    f.write(str(query))

                top_k_results = ["error-in-retrieve"]
                top_k_result_doc_ids = ["error-in-retrieve"]

        output_dr = DataRecord.from_parent(self.output_schema, parent_record=candidate)
        setattr(output_dr, self.output_attr, top_k_results)
        output_dr._evidence_file_ids = top_k_result_doc_ids

        duration_secs = time.time() - start_time
        answer = {self.output_attr: top_k_results}
        record_state = output_dr.as_dict(include_bytes=False)
        record_state["_evidence_file_ids"] = top_k_result_doc_ids

        # NOTE: right now this should be equivalent to [self.output_attr], but in the future we may
        #       want to support the RetrieveOp generating multiple fields. (Also, the function will
        #       return the full field name (as opposed to the short field name))
        generated_fields = self.get_fields_to_generate(candidate, self.input_schema, self.output_schema)

        record_op_stats = RecordOpStats(
            record_id=output_dr.id,
            record_parent_id=output_dr.parent_id,
            record_source_id=output_dr.source_id,
            record_state=record_state,
            op_id=self.get_op_id(),
            logical_op_id=self.logical_op_id,
            op_name=self.op_name(),
            time_per_record=duration_secs,
            cost_per_record=0.0,
            answer=answer,
            input_fields=self.input_schema.field_names(),
            generated_fields=generated_fields,
            fn_call_duration_secs=duration_secs,
            op_details={k: str(v) for k, v in self.get_id_params().items()},
        )

        drs = [output_dr]
        record_op_stats_lst = [record_op_stats]

        # construct record set
        record_set = DataRecordSet(drs, record_op_stats_lst)

        return record_set



from __future__ import annotations

import abc
import base64
import json
import os
import sys
from io import BytesIO
from typing import Any, Callable

import modal
import pandas as pd
from bs4 import BeautifulSoup
from papermage import Document

from palimpzest import constants
from palimpzest.core.elements.records import DataRecord
from palimpzest.core.lib.schemas import File, ImageFile, Number, PDFFile, Schema, TextFile, WebPage, XLSFile
from palimpzest.tools.pdfparser import get_text_from_pdf


# First level of abstraction
class AbstractDataSource(abc.ABC):
    """
    An AbstractDataSource is an Iterable which yields DataRecords adhering to a given schema.

    This base class must have its `__iter__` method implemented by a subclass, with each
    subclass reading data files from some real-world source (i.e. a directory, an S3 prefix,
    etc.).

    Many (if not all) DataSources should use Schemas from `palimpzest.elements.core`.
    In the future, programmers can implement their own DataSources using custom Schemas.
    """

    def __init__(self, schema: Schema) -> None:
        self._schema = schema

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(schema={self.schema})"

    def __eq__(self, __value: object) -> bool:
        return self.__dict__ == __value.__dict__

    @abc.abstractmethod
    def __len__(self) -> int: ...

    @abc.abstractmethod
    def get_item(self, idx: int) -> DataRecord: ...

    @abc.abstractmethod
    def get_size(self) -> int: ...

    @property
    def schema(self) -> Schema:
        return self._schema

    def copy(self) -> AbstractDataSource:
        raise NotImplementedError("You are calling this method from an abstract class.")

    def serialize(self) -> dict[str, Any]:
        return {"schema": self._schema.json_schema()}


class DataSource(AbstractDataSource):
    def __init__(self, schema: Schema, dataset_id: str) -> None:
        super().__init__(schema)
        self.dataset_id = dataset_id

    def universal_identifier(self):
        """Return a unique identifier for this Set."""
        return self.dataset_id

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(schema={self.schema}, dataset_id={self.dataset_id})"


# Second level of abstraction
class DirectorySource(DataSource):
    """DirectorySource returns multiple File objects from a real-world source (a directory on disk)"""

    def __init__(self, path: str, dataset_id: str, schema: Schema) -> None:
        self.filepaths = [
            os.path.join(path, filename)
            for filename in sorted(os.listdir(path))
            if os.path.isfile(os.path.join(path, filename))
        ]
        self.path = path
        super().__init__(schema, dataset_id)

    def serialize(self) -> dict[str, Any]:
        return {
            "schema": self.schema.json_schema(),
            "path": self.path,
            "source_type": "directory",
        }

    def __len__(self):
        return len(self.filepaths)

    def get_size(self):
        # Get the memory size of the files in the directory
        return sum([os.path.getsize(filepath) for filepath in self.filepaths])

    def get_item(self, idx: int):
        raise NotImplementedError("You are calling this method from an abstract class.")


class FileSource(DataSource):
    """FileSource returns a single File object from a single real-world local file"""

    def __init__(self, path: str, dataset_id: str) -> None:
        super().__init__(File, dataset_id)
        self.filepath = path

    def copy(self):
        return FileSource(self.filepath, self.dataset_id)

    def serialize(self) -> dict[str, Any]:
        return {
            "schema": self.schema.json_schema(),
            "path": self.filepath,
            "source_type": "file",
        }

    def __len__(self):
        return 1

    def get_size(self):
        # Get the memory size of the filepath
        return os.path.getsize(self.filepath)

    def get_item(self, idx: int) -> DataRecord:
        dr = DataRecord(self.schema, source_id=self.filepath)
        dr.filename = self.filepath
        with open(self.filepath, "rb") as f:
            dr.contents = f.read()

        return dr


class MemorySource(DataSource):
    """MemorySource returns multiple objects that reflect contents of an in-memory Python list"""

    def __init__(self, vals: list[int | float], dataset_id: str):
        # For the moment we assume that we are given a list of floats or ints, but someday it could be strings or something else
        super().__init__(Number, dataset_id)
        self.vals = vals

    def copy(self):
        return MemorySource(self.vals, self.dataset_id)

    def __len__(self):
        return len(self.vals)

    def get_size(self):
        return sum([sys.getsizeof(self.get_item(idx)) for idx in range(len(self))])

    def get_item(self, idx: int) -> DataRecord:
        value = self.vals[idx]
        dr = DataRecord(self.schema, source_id=idx)
        dr.value = value

        return dr


# Third level of abstraction
class HTMLFileDirectorySource(DirectorySource):
    def __init__(self, path: str, dataset_id: str) -> None:
        super().__init__(path=path, dataset_id=dataset_id, schema=WebPage)
        assert all([filename.endswith(tuple(constants.HTML_EXTENSIONS)) for filename in self.filepaths])

    def copy(self):
        return HTMLFileDirectorySource(self.path, self.dataset_id)

    def html_to_text_with_links(self, html):
        # Parse the HTML content
        soup = BeautifulSoup(html, "html.parser")

        # Find all hyperlink tags
        for a in soup.find_all("a"):
            # Check if the hyperlink tag has an 'href' attribute
            if a.has_attr("href"):
                # Replace the hyperlink with its text and URL in parentheses
                a.replace_with(f"{a.text} ({a['href']})")

        # Extract text from the modified HTML
        text = soup.get_text(separator="\n", strip=True)
        return text

    def get_item(self, idx: int) -> DataRecord:
        filepath = self.filepaths[idx]
        dr = DataRecord(self.schema, source_id=filepath)
        dr.filename = os.path.basename(filepath)
        with open(filepath) as f:
            text_content = f.read()

        html = text_content
        tokens = html.split()[: constants.MAX_HTML_ROWS]
        dr.html = " ".join(tokens)

        stripped_html = self.html_to_text_with_links(text_content)
        tokens = stripped_html.split()[: constants.MAX_HTML_ROWS]
        dr.text = " ".join(tokens)

        return dr


class ImageFileDirectorySource(DirectorySource):
    def __init__(self, path: str, dataset_id: str) -> None:
        super().__init__(path=path, dataset_id=dataset_id, schema=ImageFile)
        assert all([filename.endswith(tuple(constants.IMAGE_EXTENSIONS)) for filename in self.filepaths])

    def copy(self):
        return ImageFileDirectorySource(self.path, self.dataset_id)

    def get_item(self, idx: int) -> DataRecord:
        filepath = self.filepaths[idx]
        dr = DataRecord(self.schema, source_id=filepath)
        dr.filename = os.path.basename(filepath)
        with open(filepath, "rb") as f:
            dr.contents = base64.b64encode(f.read())
        return dr


class PDFFileDirectorySource(DirectorySource):
    def __init__(
        self,
        path: str,
        dataset_id: str,
        pdfprocessor: str = "modal",
        file_cache_dir: str = "/tmp",
    ) -> None:
        super().__init__(path=path, dataset_id=dataset_id, schema=PDFFile)
        assert all([filename.endswith(tuple(constants.PDF_EXTENSIONS)) for filename in self.filepaths])
        self.pdfprocessor = pdfprocessor
        self.file_cache_dir = file_cache_dir

    def copy(self):
        return PDFFileDirectorySource(self.path, self.dataset_id)

    def get_item(self, idx: int) -> DataRecord:
        filepath = self.filepaths[idx]
        pdf_filename = os.path.basename(filepath)
        with open(filepath, "rb") as f:
            pdf_bytes = f.read()

        if self.pdfprocessor == "modal":
            print("handling PDF processing remotely")
            remote_func = modal.Function.lookup("palimpzest.tools", "processPapermagePdf")
        else:
            remote_func = None

        # generate text_content from PDF
        if remote_func is not None:
            doc_json_str = remote_func.remote([pdf_bytes])
            docdict = json.loads(doc_json_str[0])
            doc = Document.from_json(docdict)
            text_content = ""
            for p in doc.pages:
                text_content += p.text
        else:
            text_content = get_text_from_pdf(pdf_filename, pdf_bytes, pdfprocessor=self.pdfprocessor, file_cache_dir=self.file_cache_dir)

        # construct data record
        dr = DataRecord(self.schema, source_id=filepath)
        dr.filename = pdf_filename
        dr.contents = pdf_bytes
        dr.text_contents = text_content[:15000]  # TODO Very hacky

        return dr


class TextFileDirectorySource(DirectorySource):
    def __init__(self, path: str, dataset_id: str) -> None:
        super().__init__(path=path, dataset_id=dataset_id, schema=TextFile)

    def copy(self):
        return TextFileDirectorySource(self.path, self.dataset_id)

    def get_item(self, idx: int) -> DataRecord:
        filepath = self.filepaths[idx]
        dr = DataRecord(self.schema, source_id=filepath)
        dr.filename = os.path.basename(filepath)
        with open(filepath) as f:
            dr.contents = f.read()
        return dr


class XLSFileDirectorySource(DirectorySource):
    def __init__(self, path: str, dataset_id: str) -> None:
        super().__init__(path=path, dataset_id=dataset_id, schema=XLSFile)
        assert all([filename.endswith(tuple(constants.XLS_EXTENSIONS)) for filename in self.filepaths])

    def copy(self):
        return XLSFileDirectorySource(self.path, self.dataset_id)

    def get_item(self, idx: int) -> DataRecord:
        filepath = self.filepaths[idx]
        dr = DataRecord(self.schema, source_id=filepath)
        dr.filename = os.path.basename(filepath)
        with open(filepath, "rb") as f:
            dr.contents = f.read()

        xls = pd.ExcelFile(BytesIO(dr.contents), engine="openpyxl")
        dr.number_sheets = len(xls.sheet_names)
        dr.sheet_names = xls.sheet_names
        return dr


# User-defined datasources
class UserSource(DataSource):
    """UserSource is a DataSource that is created by the user and not loaded from a file"""

    def __init__(self, schema: Schema, dataset_id: str) -> None:
        super().__init__(schema, dataset_id)

    def serialize(self) -> dict[str, Any]:
        return {
            "schema": self.schema.json_schema(),
            "source_type": "user-defined:" + self.__class__.__name__,
        }

    def __len__(self):
        raise NotImplementedError("User needs to implement this method")

    def get_size(self):
        raise NotImplementedError("User may optionally implement this method.")

    def get_item(self, idx: int) -> DataRecord:
        raise NotImplementedError("User needs to implement this method.")

    def copy(self):
        raise NotImplementedError("User needs to implement this method.")

class ValidationDataSource(UserSource):
    """
    TODO: update this class interface (and comment)
    """

    def get_val_length(self) -> int:
        raise NotImplementedError("User needs to implement this method.")

    def get_field_to_metric_fn(self) -> Callable:
        raise NotImplementedError("User needs to implement this method.")

    def get_item(self, idx: int, val: bool = False, include_label: bool = False) -> DataRecord:
        raise NotImplementedError("User needs to implement this method.")



from __future__ import annotations

from palimpzest.sets import Dataset
from typing import Any, List, Optional, Union

import palimpzest as pz

# NOTE: THINK LOW-LEVEL
# what are the abstractions that support many agent systems?
# do NOT create high-level agents (e.g. RAGAgent, ToolAgent, etc.)
# do NOT bake-in communication patterns (e.g. guardrails, summarization, etc.)
# find the general subset of operations and optimize those

# TODO: implement the following applications:
# - plain-old RAG agent (find emails in enron)
# - tool use agent ()
# - mixture of agents (see paper for example)
# - multi-agent system w/non-deterministic compute (e.g. backtrack if need be)


# class Agent:
#     def __init__(self, instruction: str):
#         self.instruction = instruction

#     # def __call__(self, input: pz.Dataset, prompt: Optional[str]=None):
#     #     prompt = (
#     #         f"You are a helpful assistant with the following instruction:\n\nInstruction {self.instruction}"
#     #         if prompt is None
#     #         else prompt
#     #     )

#     def retrieve(self, input: pz.Dataset, k: int) -> pz.Dataset:
#         assert self.resources is not None, "Need to provide a vector store for retrieval"
#         for record in input:
#             documents = self.resources.retrieve(record, k=k)
#             prompt = f"""
#               Given the input {input.__class__.__name__} and documents provided in the following context, execute the following instruction:
#               Input {record}
#               Context: {documents}
#               Instruction: {self.instruction}
#             """
#             # ... invoke LLM

#     def rerank(self, input: pz.Dataset) -> pz.Dataset:
#         for record in input:
#             prompt = f"""
#                 Given the input {input.__class__.__name__} rank the input according to the following instruction:
#                 Input {record}
#                 Instruction {self.instruction}
#             """
#             # ... invoke LLM 


class Agent:
    def __init__(self, instruction: str, schema: pz.Schema, retriever: Optional[Retriever]=None):
        self.instruction = instruction
        self.schema = schema
        self.retriever = retriever

    def __call__(self, input: Union[pz.Dataset, List[pz.Dataset]]) -> pz.Dataset:
        # handle aggregation in MoA example
        if type(input) == type([]):
            intermediates = []
            for input_ds in input:
                inter_ds = pz.Dataset(source=input_ds, schema=self.schema)
                intermediates.append(inter_ds)
            # NOTE: the Dataset construction below supposes a pz.Join logical operator
            return pz.Dataset(source=intermediates, schema=self.schema)

        return pz.Dataset(source=input, schema=self.schema, retriever=self.retriever)

#### Single-Layer Mixture of Agents for Question Answering ####
# construct schemas
class Question(pz.Schema):
    question = pz.StringField("The question which needs to be answered.")

class Answer(pz.Schema):
    answer = pz.StringField("The answer to the given question")

# construct agents
prop_agents = [Agent(instruction="Answer the given Question to the best of your ability", schema=Answer) for _ in range(5)]
agg_agent = [Agent(instruction="Summarize the provided answers into a single, concise answer", Schema=Answer)]

# construct logical plan
questions = pz.Dataset("questions", schema=Question)
prop_answers = []
for prop_agent in prop_agents:
    prop_answers = prop_agent(questions)
final_answers = agg_agent(prop_answers)

#### RAG agent for synthesizing code ####
import chromadb

class Retriever:
    def __init__(self, source: Any, collection_name: Optional[str]):
        self.source = source
        self.collection_name = collection_name

    def retrieve(self, query: str, k: int) -> List[pz.TextFile]:
        # query and k are provided by the physical operator which will ultimately invoke this method;
        # thus, k can be parameterized by the PhysicalPlanner
        files = None
        if type(self.source) == chromadb.PersistentClient:
            collection = self.source.get_or_create_collection(name=self.collection_name)
            files = collection.query(query_texts=[query], n_results=k)
        else:
            raise Exception("Not implemented yet")
            
        return files

# construct schemas
class UserInput(pz.Schema):
    user_input = pz.StringField("Text from the user describing what code they wish to edit or add to the codebase")

class CodeDiff(pz.Schema):
    code_diff = pz.StringField("Code diff to be appplied to the code base to satisfy the user's request")
    code_filename = pz.StringField("The file to apply the code diff to")
    code_start_line_number = pz.NumericField("The line at which to apply the code diff")

# construct agents
codegen_agent = Agent(
    instruction=""""
        Take the given UserInput and any related code files and produce a
        code diff. making the change(s) or addition(s) request by the user.
    """,
    retriever=Retriever(
        source=chromadb.PersistentClient(path="/path/to/vectorstore"), # could also specify source="path/to/code/"
        collection_name="codebase",
    ),
    schema=CodeDiff,
)

# construct logical plan
user_inputs = pz.Dataset("user_inputs", schema=UserInput)
code_diffs = codegen_agent(user_inputs)



#### An example of a PyTorch direction for PZ ####
import palimpzest as pz
from palimpzest.operators import convert, filter, join
from palimpzest.foundations import BaseAgent

class ProposerAgent(BaseAgent):
    def __init__(self, instruction: str):
        self.instruction = instruction

    def __call__(self, input: pz.Dataset, outputSchema: pz.Schema):
        return convert(input, desc=self.instruction, schema=outputSchema)

class AggregatorAgent(BaseAgent):
    def __init__(self, instruction: str):
        self.instruction = instruction

    def __call__(self, inputs: List[pz.Dataset], outputSchema: pz.Schema):
        agg_input = join(inputs)
        final_output = convert(agg_input, desc=self.instruction, schema=outputSchema)
        for record in final_output:
            if record.contains(self.bad_words):
                final_output = convert(agg_input, desc=self.instruction, schema=outputSchema)

        return final_output

# construct schemas
class Question(pz.Schema):
    question = pz.StringField("The question which needs to be answered.")

class Answer(pz.Schema):
    answer = pz.StringField("The answer to the given question")

# construct agents
prop_agents = [ProposerAgent(instruction="Answer the given Question to the best of your ability") for _ in range(5)]
agg_agent = [Agent(instruction="Summarize the provided answers into a single, concise answer")]

# construct logical plan
questions = pz.Dataset("questions", schema=Question)
prop_answers = []
for prop_agent in prop_agents:
    prop_answers = prop_agent(questions, outputSchema=Answer)
final_answers = agg_agent(prop_answers, outputSchema=Answer)

# execute the plan
results = pz.Execute(final_answers)


# ### defined in primitives.py ###
# def convert(input: pz.Dataset, ..., schema: pz.Schema):
#   # pz.Dataset maintains a list of the logical operations applied
#   input.ops.append(ConvertScan(input.schema, schema, ...))

#   # the list above will ultimately comprise the logical plan that we optimize

#   # then, to make the code physically function at compile time
#   for record in input:
#     for field in schema.fieldNames():
#       if isinstance(field, pz.StringField):
#         setattr(record, field, "")
#       if isinstance(field, pz.NumericField):
#         setattr(record, field, 0)
#       # ...

#   return input


# ### Defined by the user ###
# class ProposerAgent(BaseAgent):
#     def __init__(self, instruction: str):
#         self.instruction = instruction

#     def __call__(self, input: pz.Dataset, outputSchema: pz.Schema):
#         answer_records = convert(input, desc=self.instruction, schema=outputSchema)
#         for record in answer_records:
#              for bad_word in self.bad_words:
#                  record.answer = answer_record.answer.replace(bad_word, "MIT")
#         return answer_records


# # construct schemas
# class Question(pz.Schema):
#     question = pz.StringField("The question which needs to be answered.")

# class Answer(pz.Schema):
#     answer = pz.StringField("The answer to the given question")

# # construct agent
# prop_agent = ProposerAgent(instruction="answer the question")

# # construct logical plan
# questions = pz.Dataset("questions", schema=Question)
# answers = prop_agent(questions)

# # compile logical plan
# plan = pz.Compile(answers)

# # execute compiled plan (and obviously improve after getting sample data) 
# results = pz.Execute(plan)


from typing import List

from fuzzywuzzy import fuzz, process


def find_best_range(values, budget, trim_zeros=False):
    """
    Finds the consecutive range with the biggest sum within a budget.

    Args:
        values: A list of non-negative numbers.
        budget: The maximum number of consecutive elements to consider.

    Returns:
        A tuple containing the start and end indices (inclusive) of the best range,
        or None if the array is empty.
    """
    if not values:
        return None

    n = len(values)
    best_sum, best_start, current_sum, current_start = 0, 0, 0, 0

    # Iterate through the array, keeping track of current and best ranges.
    for i in range(n):
        current_sum += values[i]

        # If the current range exceeds the budget, remove elements from the beginning.
        while current_start + budget - 1 < i and current_start + budget - 1 >= 0:
            current_sum -= values[current_start]
            current_start += 1

        # Update best range if the current sum is bigger.
        if current_sum > best_sum:
            best_sum = current_sum
            best_start = current_start

    best_end = best_start + budget - 1
    print("best_start:", best_start, "best_end:", best_end)
    if trim_zeros:
        # Trim leading/trailing zeros
        while best_start >= 0 and values[best_start] == 0:
            best_start += 1

        while best_end < n and values[best_end] == 0:
            best_end -= 1
    else:
        # balance the zero entries equally on both sides
        leading_zeros = 0
        trailing_zeros = 0
        start_idx = best_start
        end_idx = best_end
        while start_idx >= 0 and values[start_idx] == 0:
            leading_zeros += 1
            start_idx += 1
        while end_idx < n and values[end_idx] == 0:
            trailing_zeros += 1
            end_idx -= 1
        half_zeros = int((leading_zeros + trailing_zeros) / 2)
        print("leading_zeros:", leading_zeros, "trailing_zeros:", trailing_zeros, "half_zeros:", half_zeros)
        best_start = best_start - half_zeros + leading_zeros
        best_end = best_end - trailing_zeros + leading_zeros + trailing_zeros - half_zeros

        if best_start < 0:
            best_end = best_end - best_start
            best_start = 0
        if best_end >= n:
            best_start = best_start - (best_end - n + 1)
            best_end = n - 1

    return best_start, best_end + 1


def get_range_from_hist(file_path, range_budget, resolution=0.001, trim_zeros=True):
    # Load data from csv file and extract he second column as values
    values = []
    with open(file_path) as file:
        for line in file:
            line = line.strip()
            values.append(int(float(line.split(",")[1])))
    index_range = 1 / resolution
    budget = int(range_budget * index_range)
    # Find the best range
    range = find_best_range(values, budget, trim_zeros=trim_zeros)
    if not range:
        raise ValueError("No range found")
    start, end = range
    print("start:", start, "end:", end, "index_range:", index_range)
    return start * 1.0 / index_range, end * 1.0 / index_range


def best_substring_match(query: str, context: str | List[str]):
    # This will extract all substrings of length equal to the query from the string
    candidates = [context[i : i + len(query)] for i in range(len(context) - len(query) + 1)]

    # Find the best match among the candidates
    ret = process.extractOne(query, candidates, scorer=fuzz.ratio)
    if ret is None:
        return None

    best_match, score = ret
    positions = [can == best_match for can in candidates]
    start = positions.index(True)
    end = start + len(query)
    # print("best match:", best_match, "score:", score, "start:", start, "end:", end)
    # print("-------", string[start:end])
    return start, end
